/*
code.cc -- control flow handling for geo
Copyright (C) Zsban Ambrus 2003, 2004

This source defines the yylex function that yyparse will call.
yylex interprets tokens stored in the token list by the lexer sequentially,
and does special things for handling jump constructs.

Csehorszg nyugati rszn rendkvli fagy tombol, de nem jogers, a lakossg
az id enyhtsrt fellebbezett.

*/


#include "all.hh"
#include "parse.tab.hh"
#include <list>
#include <iostream>
#include <stack>

using std::cerr;



// a tokeneket sosem szabaditjuk fel!!!

static std::list<token *> *code_list;

static code_iterator_t code_iterator;
static code_iterator_t code_end;

static int code_buf;


/* dynamic scope: internal data */


class localize_t {
	UNLEAK
	public:
	localize_t (Lvalue *);
	~localize_t (void);
	private:
	Lvalue *const ptr;
	Value *const sav;
	};

/* 
The stack of block_t is used in two ways.  One is with the block_begin,
block_end functions to handle () blocks.  The other is the calling of
lambdas, with block_call and block_end.  The iterator cont is only used
in the second case: it stores the point from where the lambda is called
from, and where block_leave will jump back.  The parser always knows
which case there is, but we don't use its info here.  In both cases,
code points to the beginning of the block, to which the parser can jump
back with the block_redo function.  Local variables are stored in the
stack.top(), used through the function block_local.
*/

class block_t {
	UNLEAK
	public:
	block_t (void);
	virtual ~block_t (void);
	const code_iterator_t code;
	std::list<localize_t *> locals;
	protected:
	block_t (const code_iterator_t &cd);
	};

class block_lambda_t:
	public block_t {
	public:
	block_lambda_t (const code_iterator_t &);
	~block_lambda_t (void);
	const code_iterator_t cont;
	};

static std::stack<block_t *> stack;



/* lexer */

void init_parse (std::list<token *> *tokens) {
	code_list= tokens;
	code_iterator= code_list->begin ();
	code_end= code_list->end ();
	code_buf= 0;
	}

int valueof (token *t) {
	if (dynamic_cast<token_int *>(t)) {
		yylval.v= new Val_int (dynamic_cast<token_int *>(t)->val);
		return T_INT;
		}
	if (dynamic_cast<token_punct *>(t)) {
		yylval.v= 0;
		return dynamic_cast<token_punct *>(t)->chr;
		}
	if (dynamic_cast<token_sym *>(t)) {
		yylval.var= new Lval_var (dynamic_cast<token_sym *>(t));
		return T_SYM;
		}
	cerr <<"internal error: unknown type of token\n";
	exit (1);
	}

int yylex (void) {
	int b= code_buf;
	code_buf= 0;
	if (b)
		return b;
	if (code_iterator==code_end)
		return 0;
	token *t= *code_iterator++;
	return valueof (t);
	}

void unlex (int c) {
	if (code_buf) {
		cerr <<"internal error: double unlex\n";
		exit (1);
		}
	code_buf= c;
	}
	

/* dynamic scope: impl */

block_t::block_t (void):
	code (code_iterator) {
	}

// this function is only for block_lambda_t::block_local_t()'s use only!
block_t::block_t (const code_iterator_t &a):
	code (a) { 
	}

static void done_local (localize_t *t) {
	delete t;
	}

block_t::~block_t (void) {
	for_each (locals.begin (), locals.end (), &done_local);
	}

localize_t::localize_t (Lvalue *p):
	ptr (p),
	sav (p->value ()) {
	}

localize_t::~localize_t (void) {
	delete ptr->set (sav);
	delete ptr;
	}

void block_begin (void) {
	stack.push (new block_t);
	}

void block_end (void) {
	if (stack.empty ()) {
		cerr <<"internal error: no block when trying to exit block\n";
		exit (1);
		}
	delete stack.top ();
	stack.pop ();
	}

void block_redo (void) {
	if (stack.empty ()) {
		cerr <<"internal error: no block when trying to redo\n";
		exit (1);
		}
	code_iterator= stack.top ()->code;
	}

void block_local (Lvalue *l) {
	stack.top ()->locals.push_front (new localize_t (l));
	}


/* lambdas (but no 1st class functions) */

Val_lambda::Val_lambda (void):
	pos (code_iterator) {
	}
	
Val_lambda::~Val_lambda (void) {
	}

Val_lambda::Val_lambda (Val_lambda &f):
	pos (f.pos) {
	}
	
Val_lambda *Val_lambda::copy (void) {
	return new Val_lambda (*this);
	}

block_lambda_t::block_lambda_t (const code_iterator_t &g):
	block_t (g), cont (code_iterator) {
	code_iterator= g;
	}

block_lambda_t::~block_lambda_t (void) {
	code_iterator= cont;
	}

void block_call (Val_lambda *l) {
	stack.push (new block_lambda_t (l->pos));
	}



//END
